#ifdef HAVE_CONFIG_H
#include	"config.h"
#endif

#include	<stdio.h>
#ifdef HAVE_STDLIB_H
#include	<stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include	<string.h>
#endif

#ifdef HAVE_KSTAT_H
#include	<kstat.h>
#endif
#ifdef HAVE_SYS_SYSINFO_H
#include	<sys/sysinfo.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include	<sys/types.h>
#endif
#ifdef HAVE_SYS_SYSCTL_H
#include	<sys/sysctl.h>
#endif
#ifdef HAVE_MACH_O_ARCH_H
#include	<mach-o/arch.h>
#include	<sys/vmmeter.h>
#include	<mach/mach.h>
#include	<mach/host_info.h>
#endif
#ifdef HAVE_SYS_SYSTEMINFO_H
#include	<sys/systeminfo.h>
#endif

#ifdef HAVE_SYS_PARAM_H
#include	<sys/param.h>
#endif

#ifdef HAVE_W32API_BASETYPS_H
#include	<w32api/basetyps.h>
#endif
#include <jni.h>

#include "jp_co_sds_0005fcorp_util_device_DeviceManager.h"

struct cpuinfo {
	char	name[128];
	char	vendor[128];
	char	speed[128];
	int		num;
	char	kernel[256];
};

struct cpustat {
	long long	user;
	long long	nice;
	long long	system;
	long long	idle;
	long long	total;
};
static int init = 0;
static struct cpustat stat_s, stat_e;

struct cpuload {
	double	avg1min;
	double	avg5min;
	double	avg15min;
};

void Throw( JNIEnv *env, const char *name, const char *msg);
static void copyData( char *dst, char *src );
static void readCpuinfo( JNIEnv *env, struct cpuinfo *info );
static void readVersion( JNIEnv *env, struct cpuinfo *info );
static void readLoadavg( JNIEnv *env, struct cpuload *info );
static void readStat( JNIEnv *env, struct cpustat *info, struct cpuinfo *cpuinfo );
static void getLoadavg( JNIEnv *env, struct cpuload *info );
static void readKstat( JNIEnv *env, struct cpuinfo *cpuinfo,
	struct cpuload *cpuload );
static void readKstat_stat( JNIEnv *env, struct cpustat *stat_s );
static void readSysctl( JNIEnv *env, struct cpuinfo *cpuinfo,
	struct cpuload *cpuload );
static void readSysctl_stat( JNIEnv *env, struct cpustat *stat );

#ifdef HAVE_LIBKSTAT
void initKstat();
long long getKStatNumber( JNIEnv *env, kstat_ctl_t *kc, char *moduleName, 
					char *recordName, char *fieldName);
char *getKStatString( JNIEnv *env, kstat_ctl_t *kc, char *moduleName, 
					char *recordName, char *fieldName);
#endif
void getSysctlString( int n1, int n2, char *buf, size_t len );
int getSysctlInt( int n1, int n2 );

/*
 * Class:     jp_co_sds_0005fcorp_util_device_DeviceManager
 * Method:    getCPUStatus
 * Signature: ()Ljp/co/sds_corp/util/device/CPUStatus;
 */
JNIEXPORT jobject JNICALL Java_jp_co_sds_1corp_util_device_DeviceManager_getCPUStatus
	( JNIEnv *env, jclass cls, jint sec )
{
	int		nRet = -1;
	jfieldID    field;
	jstring		str;
	jclass 		retCls;
	jobject ret;
	struct cpuinfo cpuinfo;
	struct cpuload cpuload;
	double total;
	double dU, dN, dS, dI;

	retCls = (*env)->FindClass( env, "jp/co/sds_corp/util/device/CPUStatus" );
	ret = (*env)->AllocObject( env, retCls );

	memset( &cpuinfo, 0, sizeof( struct cpuinfo ));
	memset( &cpuload, 0, sizeof( struct cpuload ));
//	memset( &stat_s, 0, sizeof( struct cpustat ));
	memset( &stat_e, 0, sizeof( struct cpustat ));

#ifdef HAVE_LIBKSTAT
	readKstat( env, &cpuinfo, &cpuload );
	if ( init == 0 ) {
		readKstat_stat( env, &stat_s );
		sleep( sec );
		init = 1;
	}
	readKstat_stat( env, &stat_e );
#else
#ifdef HAVE_PROC_CPUINFO
	readCpuinfo( env, &cpuinfo );
#ifdef HAVE_PROC_VERSION
	readVersion( env, &cpuinfo );
#endif
#else
#ifdef HAVE_SYSCTL
	readSysctl( env, &cpuinfo, &cpuload );
	if ( init == 0 ) {
		readSysctl_stat( env, &stat_s );
		sleep( sec );
		init = 1;
	}
	readSysctl_stat( env, &stat_e );
#endif // SYSCTL
#endif

#ifdef HAVE_PROC_STAT
	if ( init == 0 ) {
		readStat( env, &stat_s, &cpuinfo );
		sleep( sec );
		init = 1;
	}
	readStat( env, &stat_e, &cpuinfo );
#endif

#endif // LIBKSTAT

#ifdef HAVE_GETLOADAVG
	getLoadavg( env, &cpuload );
#else
#ifdef HAVE_PROC_LOADAVG
	readLoadavg( env, &cpuload );
#endif
#endif

	str = (*env)->NewStringUTF( env, cpuinfo.name );
	field = (*env)->GetFieldID( env, retCls, "name", "Ljava/lang/String;" );
	(*env)->SetObjectField( env, ret, field, str );
	str = (*env)->NewStringUTF( env, cpuinfo.vendor );
	field = (*env)->GetFieldID( env, retCls, "vendor", "Ljava/lang/String;" );
	(*env)->SetObjectField( env, ret, field, str );
	str = (*env)->NewStringUTF( env, cpuinfo.speed );
	field = (*env)->GetFieldID( env, retCls, "speed", "Ljava/lang/String;" );
	(*env)->SetObjectField( env, ret, field, str );
	str = (*env)->NewStringUTF( env, cpuinfo.kernel );
	field = (*env)->GetFieldID( env, retCls, "kernel", "Ljava/lang/String;" );
	(*env)->SetObjectField( env, ret, field, str );

	field = (*env)->GetFieldID( env, retCls, "avg1min", "D" );
	(*env)->SetDoubleField( env, ret, field, cpuload.avg1min );
	field = (*env)->GetFieldID( env, retCls, "avg5min", "D" );
	(*env)->SetDoubleField( env, ret, field, cpuload.avg5min );
	field = (*env)->GetFieldID( env, retCls, "avg15min", "D" );
	(*env)->SetDoubleField( env, ret, field, cpuload.avg15min );

	field = (*env)->GetFieldID( env, retCls, "num", "I" );
	(*env)->SetIntField( env, ret, field, cpuinfo.num );

	total = (double )stat_e.total - (double )stat_s.total;
	dU = (stat_e.user - stat_s.user) / (double )total * 100.00;
	dN = (stat_e.nice - stat_s.nice) / (double )total * 100.00;
	dS = (stat_e.system - stat_s.system) / (double )total * 100.00;
	dI = 100.00 - dU - dS - dN;
#ifdef DEBUG
printf( "%lld %lld %lld / (%lld - %lld) %lld => %f %f %f %f\n",
		(stat_e.user - stat_s.user),
		(stat_e.nice - stat_s.nice),
		(stat_e.system - stat_s.system),
		stat_e.total, stat_s.total,
		total, dU, dN, dS, dI );
#endif
	memcpy( &stat_s, &stat_e, sizeof( struct cpustat ));

	field = (*env)->GetFieldID( env, retCls, "userTime", "D" );
	(*env)->SetDoubleField( env, ret, field, dU );
	field = (*env)->GetFieldID( env, retCls, "niceTime", "D" );
	(*env)->SetDoubleField( env, ret, field, dN );
	field = (*env)->GetFieldID( env, retCls, "systemTime", "D" );
	(*env)->SetDoubleField( env, ret, field, dS );
	field = (*env)->GetFieldID( env, retCls, "idleTime", "D" );
	(*env)->SetDoubleField( env, ret, field, dI );
	field = (*env)->GetFieldID( env, retCls, "totalTime", "D" );
	(*env)->SetDoubleField( env, ret, field, 100.0 );

	return ret;
}

static void copyData( char *dst, char *src ) {
	int nPos;
	for ( nPos =0; nPos < strlen( src ); nPos++ ) {
		if ( src[nPos] == ':' ) {
			break;
		}
	}
	strcpy( dst, src + nPos + 2 );
	if ( dst[strlen(dst)-1] == '\n' ) {
		dst[strlen(dst)-1] = '\0';
	}
}

static void readCpuinfo( JNIEnv *env, struct cpuinfo *info ) {
	char	*file;
	FILE	*pFp;
	char	buf[1024];
	char	*key;

	file = "/proc/cpuinfo";
	pFp = fopen( file, "r" );
	if ( pFp == NULL ) {
		Throw( env, "java/io/FileNotFoundException", file );
		return;
	}

	for ( ;; ) {
		if ( fgets( buf, sizeof( buf ), pFp ) == NULL ) {
			break;
		}
		key = "model name";
		if ( memcmp( buf, key, strlen( key )) == 0 ) {
			copyData( info->name, buf );
		}
		key = "vendor_id";
		if ( memcmp( buf, key, strlen( key )) == 0 ) {
			copyData( info->vendor, buf );
		}
		key = "cpu MHz";
		if ( memcmp( buf, key, strlen( key )) == 0 ) {
			copyData( info->speed, buf );
		}
	}
	fclose( pFp );
}

static void readVersion( JNIEnv *env, struct cpuinfo *info ) {
	char	*file;
	FILE	*pFp;
	char	buf[1024];
	char	*key;

	file = "/proc/version";
	pFp = fopen( file, "r" );
	if ( pFp == NULL ) {
		Throw( env, "java/io/FileNotFoundException", file );
		return;
	}

	if ( fgets( buf, sizeof( buf ), pFp ) == NULL ) {
		fclose( pFp );
		Throw( env, "java/io/IOException", file );
		return;
	}
	if ( buf[strlen(buf)-1] == '\n' ) {
		buf[strlen(buf)-1] = '\0';
	}
	strcpy( info->kernel, buf );
	fclose( pFp );
}

#ifdef HAVE_GETLOADAVG
static void getLoadavg( JNIEnv *env, struct cpuload *info ) {
	double	avg[3];
	int		nRet;

	nRet = getloadavg( avg, 3 );
	if ( nRet < 0 ) {
		Throw( env, "java/lang/IllegalStateException", "" );
		return;
	}
	info->avg1min = avg[0];
	info->avg5min = avg[1];
	info->avg15min = avg[2];
}
#endif

static void readLoadavg( JNIEnv *env, struct cpuload *info ) {
	char	*file;
	FILE	*pFp;
	char	buf[1024];

	file = "/proc/loadavg";
	pFp = fopen( file, "r" );
	if ( pFp == NULL ) {
		Throw( env, "java/io/FileNotFoundException", file );
		return;
	}

	for ( ;; ) {
		int nPos;
		char *p;

		if ( fgets( buf, sizeof( buf ), pFp ) == NULL ) {
			break;
		}
		nPos = 0;
		p = strtok( buf, " " );
		for ( ;; ) {
			if ( p[strlen(p)-1] == '\n' ) {
				p[strlen(p)-1] = '\0';
			}
			switch ( nPos ) {
			case 0: info->avg1min = atof( p ); break;
			case 1: info->avg5min = atof( p ); break;
			case 2: info->avg15min = atof( p ); break;
			}
			p = strtok( NULL, " " );
			if ( p == NULL ) {
				break;
			}
			nPos++;
		}
	}
	fclose( pFp );
}

static void readStat( JNIEnv *env, struct cpustat *info, struct cpuinfo *cpuinfo ) {
	char	*file;
	FILE	*pFp;
	char	buf[1024];
	char	*key;
	int		nCpu = -1;

	file = "/proc/stat";
	pFp = fopen( file, "r" );
	if ( pFp == NULL ) {
		Throw( env, "java/io/FileNotFoundException", file );
		return;
	}

	for ( ;; ) {
		if ( fgets( buf, sizeof( buf ), pFp ) == NULL ) {
			break;
		}
		key = "cpu";
		if ( memcmp( buf, key, strlen( key )) == 0 ) {
			nCpu++;
		}
		key = "cpu ";
		if ( memcmp( buf, key, strlen( key )) == 0 ) {
			int nPos = 0;
			char *p;
			p = strtok( buf, " " );
			for ( ;; ) {
				if ( p[strlen(p)-1] == '\n' ) {
					p[strlen(p)-1] = '\0';
				}
				switch ( nPos ) {
				case 1: info->user = atoll( p ); break;
				case 2: info->nice = atoll( p ); break;
				case 3: info->system = atoll( p ); break;
				case 4: info->idle = atoll( p ); break;
				}
				p = strtok( NULL, " " );
				if ( p == NULL ) {
					break;
				}
				nPos++;
			}
		}
	}
	fclose( pFp );

	info->total = info->user + info->nice + info->system + info->idle;
	cpuinfo->num = nCpu;
}

#ifdef HAVE_LIBKSTAT
/* Fetch numerical statistic from kernel */
long long getKStatNumber( JNIEnv *env, kstat_ctl_t *kc, char *moduleName, 
					char *recordName, char *fieldName)
{
	kstat_t *ksp;
	kstat_named_t *kstatFields;
	long long value;
	int i;
	char	tgt[256];

	strcpy( tgt, moduleName );
	strcat( tgt, ":" );
	strcat( tgt, recordName );
	ksp = kstat_lookup( kc, moduleName, -1, recordName );
	if ( ksp == NULL ) {
		Throw( env, "java/io/IOException", tgt );
		return -1;
	}

	if ( kstat_read( kc, ksp, NULL ) < 0) {
		Throw( env, "java/io/IOException", tgt );
		return -1;
	}

	kstatFields = KSTAT_NAMED_PTR( ksp );

	for ( i = 0; i < ksp->ks_ndata; i++ ) {
		if ( strcmp( kstatFields[i].name, fieldName ) == 0) {
			switch( kstatFields[i].data_type ) {
			case KSTAT_DATA_INT32:
				value = kstatFields[i].value.i32;
				break;
			case KSTAT_DATA_UINT32:
				value = kstatFields[i].value.ui32;
				break;
			case KSTAT_DATA_INT64:
				value = kstatFields[i].value.i64;
				break;
			case KSTAT_DATA_UINT64:
				value = kstatFields[i].value.ui64;
				break;
			default:
				strcat( tgt, ":" );
				strcat( tgt, fieldName );
				Throw( env, "java/lang/IllegalArgumentException", tgt );
				value = -1;
			}
			return value;
		}
	}
	strcat( tgt, ":" );
	strcat( tgt, fieldName );
	Throw( env, "java/io/IOException", tgt );
	return -1;
}

/* Fetch string statistic from kernel */
char *getKStatString( JNIEnv *env, kstat_ctl_t *kc, char *moduleName, 
					char *recordName, char *fieldName)
{
	kstat_t *ksp;
	kstat_named_t *kstatFields;
	char *value;
	int i;
	char	tgt[256];
	char	msg[1024];

	strcpy( tgt, moduleName );
	strcat( tgt, ":" );
	strcat( tgt, recordName );
	ksp = kstat_lookup( kc, moduleName, -1, recordName );
	if ( ksp == NULL ) {
		Throw( env, "java/io/IOException", tgt );
		return NULL;
	}

	if ( kstat_read( kc, ksp, NULL ) < 0 ) {
		Throw( env, "java/io/Exception", tgt );
		return NULL;
	}

	kstatFields = KSTAT_NAMED_PTR( ksp );

	for ( i = 0; i < ksp->ks_ndata; i++ ) {
		if ( strcmp( kstatFields[i].name, fieldName ) == 0 ) {
			switch( kstatFields[i].data_type ) {
			case KSTAT_DATA_CHAR:
				value = kstatFields[i].value.c;
				break;
			case KSTAT_DATA_STRING:
				value = KSTAT_NAMED_STR_PTR( &kstatFields[i] );
				break;
			default:
				sprintf( msg, "%s:%s=%d", tgt, fieldName,
					kstatFields[i].data_type );
				if ( kstatFields[i].value.c != NULL ) {
					strcat( msg, "[" );
					strcat( msg, kstatFields[i].value.c );
				}
				Throw( env, "java/lang/IllegalArgumentException", msg );
				value = NULL;
			}
			return value;
		}
	}
	strcat( tgt, ":" );
	strcat( tgt, fieldName );
	Throw( env, "java/io/IOException", tgt );
	return NULL;
}

int			nInit = 0;
kstat_ctl_t *kc;

void initKstat() {
	if ( nInit == 0 ) {
		kc = kstat_open();
		nInit = 1;
	}
}

static void readKstat( JNIEnv *env, struct cpuinfo *cpuinfo, struct cpuload *cpuload )
{
	long	value;
	char	*str, *type;
	char	buf[128];
	char	sys[128], rel[128], ver[128], mach[128], arch[128], plat[128];

	initKstat();
	/* Fetch CPU info */
	type = getKStatString( env, kc, "cpu_info", "cpu_info0", "cpu_type");
	str = getKStatString( env, kc, "cpu_info", "cpu_info0", "implementation");
	sprintf( cpuinfo->name, "%s(%s)", str, type );
	value = getKStatNumber( env, kc, "cpu_info", "cpu_info0", "clock_MHz");
	sprintf( cpuinfo->speed, "%ld", value );
	str = getKStatString( env, kc, "cpu_info", "cpu_info0", "state");
	value = getKStatNumber( env, kc, "unix", "system_misc", "ncpus");
	cpuinfo->num = value;
#ifdef HAVE_SYS_SYSTEMINFO_H
	sysinfo( SI_HW_PROVIDER, buf, sizeof( buf ));
	strcpy( cpuinfo->vendor, buf );

	sysinfo( SI_SYSNAME, sys, sizeof( sys ));
	sysinfo( SI_RELEASE, rel, sizeof( rel ));
	sysinfo( SI_VERSION, ver, sizeof( ver ));
	sysinfo( SI_MACHINE, mach, sizeof( mach ));
	sysinfo( SI_ARCHITECTURE, arch, sizeof( arch ));
	sysinfo( SI_PLATFORM, plat, sizeof( plat ));
	sprintf( cpuinfo->kernel, "%s %s %s %s %s %s",
							sys, rel, ver, mach, arch, plat );
#endif

	/* Fetch the current load averages */
	value = getKStatNumber( env, kc, "unix", "system_misc", "avenrun_1min");
	cpuload->avg1min = value / FSCALE;
	value = getKStatNumber( env, kc, "unix", "system_misc", "avenrun_5min");
	cpuload->avg5min = value / FSCALE;
	value = getKStatNumber( env, kc, "unix", "system_misc", "avenrun_15min");
	cpuload->avg15min = value / FSCALE;
}
static void readKstat_stat( JNIEnv *env, struct cpustat *stat )
{
	long	value;
	char	*str;

	initKstat();

	kstat_t		*ksp;
	cpu_stat_t	cpu;
	ksp = kstat_lookup( kc, "cpu_stat", 0, "cpu_stat0" );
	if ( ksp == NULL ) {
		perror( "kstat_lookup" );
		return;
	}
	if ( kstat_read( kc, ksp, &cpu ) == -1 ) {
		perror( "kstat_read" );
		return;
	}
	stat->system = cpu.cpu_sysinfo.cpu[CPU_KERNEL];
	stat->user = cpu.cpu_sysinfo.cpu[CPU_USER];
	stat->nice = 0;
	long idle = cpu.cpu_sysinfo.cpu[CPU_IDLE] + cpu.cpu_sysinfo.cpu[CPU_WAIT];
	stat->total = stat->system + stat->user + idle;
}
#endif

#if defined(HAVE_SYSCTL) && !defined(HAVE_PROC_CPUINFO)
void getSysctlString( int n1, int n2, char *buf, size_t len ) {
	int mib[2];

	mib[0] = n1;
	mib[1] = n2;
	sysctl( mib, 2, buf, &len, NULL, 0 );
}

int getSysctlInt( int n1, int n2 ) {
	int mib[2];
	size_t len;
	int	num;

	mib[0] = n1;
	mib[1] = n2;
	len = sizeof( int );
	sysctl( mib, 2, &num, &len, NULL, 0 );

	return num;
}

static void readSysctl( JNIEnv *env, struct cpuinfo *cpuinfo, struct cpuload *cpuload )
{
	char	*p;
	int		val;
	const NXArchInfo *info;
	char	buf[1024];

	val = getSysctlInt( CTL_HW, HW_NCPU );
	cpuinfo->num = val;
	val = getSysctlInt( CTL_HW, HW_CPU_FREQ );
	sprintf( cpuinfo->speed, "%d", val / 1000000 );
	getSysctlString( CTL_KERN, KERN_VERSION, buf, sizeof( buf ) );
	strcpy( cpuinfo->kernel, buf );

	info = NXGetLocalArchInfo();
	strcpy( cpuinfo->name, info->description );
}

#if 0
void printSysctl() {
	int		val;
	int mib[2];
	size_t len = 2;
	char	buf[1024];

	getSysctlString( CTL_HW, HW_MACHINE, buf, sizeof( buf ) );
	printf( "MACHINE=%s\n", buf );
	getSysctlString( CTL_HW, HW_MODEL, buf, sizeof( buf ) );
	printf( "MODEL=%s\n", buf );
	getSysctlString( CTL_HW, HW_MACHINE_ARCH, buf, sizeof( buf ) );
	printf( "MACHINE_ARCH=%s\n", buf );

	sysctlnametomib( "hw.cputype", mib, &len );
	val = getSysctlInt( mib[0], mib[1] );
	printf( "cputype=%d %s\n", val );

	sysctlnametomib( "hw.cpusubtype", mib, &len );
	val = getSysctlInt( mib[0], mib[1] );
	printf( "cpusubtype=%d\n", val );

}
#endif

static void readSysctl_stat( JNIEnv *env, struct cpustat *stat )
{
	int i;
	unsigned int count = HOST_CPU_LOAD_INFO_COUNT;
	host_cpu_load_info_data_t cpuload;
	long cp_time[CPU_STATE_MAX];

    host_statistics(mach_host_self(), HOST_CPU_LOAD_INFO,
					(host_info_t)&cpuload, &count);
	for (i = 0; i < CPU_STATE_MAX; i++) {
		cp_time[i] = cpuload.cpu_ticks[i];
	}
	stat->user = cp_time[CPU_STATE_USER];
	stat->system = cp_time[CPU_STATE_SYSTEM];
	stat->nice = cp_time[CPU_STATE_NICE];
	stat->total = cp_time[CPU_STATE_IDLE]
				+ stat->user + stat->system + stat->nice;

}
#endif

