#ifdef HAVE_CONFIG_H
#include	"config.h"
#endif

#include	<stdio.h>
#ifdef HAVE_STRING_H
#include	<string.h>
#endif

#include	<sys/types.h>
#ifdef HAVE_SYS_STATFS_H
#include	<sys/statfs.h>
#endif
#ifdef HAVE_SYS_STATVFS_H
#include	<sys/statvfs.h>
#endif
#ifdef HAVE_SYS_MNTTAB_H
#include	<sys/mnttab.h>
#endif
#ifdef HAVE_SYS_VFS_H
#include	<sys/vfs.h>
#endif
#ifdef HAVE_MNTENT_H
#include	<mntent.h>
#endif
#ifdef HAVE_SYS_MOUNT_H
#include	<sys/mount.h>
#endif
#ifdef HAVE_SYS_PARAM_H
#include	<sys/param.h>
#endif
#include	<jni.h>
#include	"jp_co_sds_0005fcorp_util_device_DeviceManager.h"

struct driveinfo {
	char	name[256];
	char	dev[256];
	char	type[256];
	long long	lBlock;
	long long	lMax;
	long long	lUsed;
	long long	lFree;
	long long	lAvail;
};

void Throw( JNIEnv *env, const char *name, const char *msg);
int getDriveInfo( JNIEnv *env, const char *pName, struct driveinfo *info );
int getDeviceInfo( JNIEnv *env, const char *pName, struct driveinfo *info );

static int countFs( JNIEnv *env );
static void startList( void );
static int list( JNIEnv *env, struct driveinfo *info );
static void endList( void );
static void setInfo( JNIEnv *env, jclass cls, jobject obj,
					struct driveinfo *info );

/*
 * Class:     jp_co_sds_0005fcorp_util_device_DeviceManager
 * Method:    getAllDriveInfo
 * Signature: ()[Ljp/co/sds_corp/util/device/DriveInfo;
 */
JNIEXPORT jobjectArray JNICALL Java_jp_co_sds_1corp_util_device_DeviceManager_getAllDriveInfo
  (JNIEnv *env, jclass cls)
{
	int		nRet;
	int		i, nNum;
	jobjectArray	ret;
	jclass			objCls;
	struct driveinfo	info[128];

	objCls = (*env)->FindClass( env, "jp/co/sds_corp/util/device/DriveInfo" );
	nNum = countFs( env );
	if ( nNum < 0 ) {
		return NULL;
	}

	memset( info, 0, sizeof( info ));
	ret = (*env)->NewObjectArray( env, nNum, objCls, NULL );
	startList();
	for ( i = 0; i < nNum; i++ ) {
		jobject obj;

		nRet = list( env, &info[i] );
		if ( nRet < 0 ) {
			return NULL;
		}
		obj = (*env)->AllocObject( env, objCls );
		(*env)->SetObjectArrayElement( env, ret, i, obj );

#ifdef DEBUG
printf( "%s\n", info[i].name );
#endif
		nRet = getDriveInfo( env, info[i].name, &info[i] );
		if ( nRet < 0 ) {
			return NULL;
		}
		setInfo( env, objCls, obj, &info[i] );
	}
	endList();

	return ret;
}

#if defined( HAVE_GETFSSTAT )
struct statfs buf[128];
int	nLoop;

static int countFs( JNIEnv *env ) {
	int		nNum = 0;

	nNum = getfsstat( buf, sizeof( buf ), MNT_WAIT );
	if ( nNum <= 0 ) {
		Throw( env, "java/io/IOException", "" );
		return -1;
	}

	return nNum;
}
static int list( JNIEnv *env, struct driveinfo *info ) {
	strcpy( info->name, buf[nLoop].f_mntonname );
	nLoop++;
	return 0;
}
static void startList() {
	nLoop = 0;
}
static void endList() {
}
int getDeviceInfo( JNIEnv *env, const char *pName, struct driveinfo *info ) {
	int		nNum;
	int		nRet = -1;
	int		i;
	nNum = getfsstat( buf, sizeof( buf ), MNT_WAIT );
	for ( i = 0; i < nNum; i++ ) {
		if ( strcmp( buf[i].f_mntonname, pName ) == 0 ) {
			strcpy( info->dev, buf[i].f_mntfromname );
			strcpy( info->type, buf[i].f_fstypename );
			nRet = 0;
			break;
		}
	}
	return nRet;
}

#elif defined( HAVE_SETMNTENT )
const char	*file = "/etc/mtab";
FILE	*pFp;
struct mntent	*pMnt;

static int countFs( JNIEnv *env ) {
	int		nNum = 0;
	FILE	*pFp;

	pFp = setmntent( file, "r" );
	if ( pFp == NULL ) {
		Throw( env, "java/io/IOException", file );
		return -1;
	}
	for ( ;; ) {
		pMnt = getmntent( pFp );
		if ( pMnt == NULL ) {
			break;
		}
		nNum++;
	}
	endmntent( pFp );
	return nNum;
}
static int list( JNIEnv *env, struct driveinfo *info ) {
	pMnt = getmntent( pFp );
	if ( pMnt == NULL ) {
		Throw( env, "java/io/IOException", file );
		return -1;
	}
	strcpy( info->name, pMnt->mnt_dir );
	strcpy( info->dev, pMnt->mnt_fsname );
	strcpy( info->type, pMnt->mnt_type );
	return 0;
}
static void startList() {
	pFp = setmntent( file, "r" );
}
static void endList() {
	endmntent( pFp );
}
int getDeviceInfo( JNIEnv *env, const char *pName, struct driveinfo *info ) {
	FILE	*pFp;
	int		nRet = -1;

	pFp = setmntent( file, "r" );
	if ( pFp == NULL ) {
		Throw( env, "java/io/IOException", file );
		return -1;
	}
	for ( ;; ) {
		pMnt = getmntent( pFp );
		if ( pMnt == NULL ) {
			break;
		}
		if ( strcmp( pMnt->mnt_dir, pName ) == 0 ) {
			strcpy( info->dev, pMnt->mnt_fsname );
			strcpy( info->type, pMnt->mnt_type );
			nRet = 0;
			break;
		}
	}
	endmntent( pFp );
	return nRet;
}

#elif defined( HAVE_GETMNTENT )
const char	*file = "/etc/mnttab";
FILE	*pFp;
struct mnttab	mnt;
static int countFs( JNIEnv *env ) {
	int		nNum = 0;
	FILE	*pFp;
	int		nRet;

	pFp = fopen( file, "r" );
	if ( pFp == NULL ) {
		Throw( env, "java/io/FileNotFoundException", file );
		return -1;
	}
	for ( ;; ) {
		nRet = getmntent( pFp, &mnt );
		if ( nRet < 0 ) {
			break;
		}
		nNum++;
	}
	fclose( pFp );
	return nNum;
}
static void startList() {
	pFp = fopen( file, "r" );
}
static int list( JNIEnv *env, struct driveinfo *info ) {
	int		nRet;
	nRet = getmntent( pFp, &mnt );
	if ( nRet < 0 ) {
		Throw( env, "java/io/FileNotFoundException", file );
		return -1;
	}
	strcpy( info->name, mnt.mnt_mountp );
	strcpy( info->dev, mnt.mnt_special );
	strcpy( info->type, mnt.mnt_fstype );
	return 0;
}
static void endList() {
	fclose( pFp );
}
int getDeviceInfo( JNIEnv *env, const char *pName, struct driveinfo *info ) {
	FILE	*pFp;
	int		ret;
	int		nRet = -1;

	pFp = fopen( file, "r" );
	if ( pFp == NULL ) {
		Throw( env, "java/io/FileNotFoundException", file );
		return -1;
	}
	for ( ;; ) {
		ret = getmntent( pFp, &mnt );
		if ( ret < 0 ) {
			break;
		}
		if ( strcmp( mnt.mnt_mountp, pName ) == 0 ) {
			strcpy( info->dev, mnt.mnt_special );
			strcpy( info->type, mnt.mnt_fstype );
			nRet = 0;
			break;
		}
	}
	fclose( pFp );
	return nRet;
}

#endif

static void setInfo( JNIEnv *env, jclass cls, jobject obj,
					struct driveinfo *info )
{
	jfieldID	field;
	jstring rname;
	jstring dev;
	jstring type;

	field = (*env)->GetFieldID( env, cls, "name", "Ljava/lang/String;" );
	rname = (*env)->NewStringUTF( env, info->name );
	(*env)->SetObjectField( env, obj, field, rname );

	field = (*env)->GetFieldID( env, cls, "deviceName", "Ljava/lang/String;" );
	dev = (*env)->NewStringUTF( env, info->dev );
	(*env)->SetObjectField( env, obj, field, dev );

	field = (*env)->GetFieldID( env, cls, "fileSystem", "Ljava/lang/String;" );
	type = (*env)->NewStringUTF( env, info->type );
	(*env)->SetObjectField( env, obj, field, type );

	field = (*env)->GetFieldID( env, cls, "blockSize", "J" );
	(*env)->SetLongField( env, obj, field, info->lBlock );

	field = (*env)->GetFieldID( env, cls, "maxSize", "J" );
	(*env)->SetLongField( env, obj, field, info->lMax );

	field = (*env)->GetFieldID( env, cls, "freeSize", "J" );
	(*env)->SetLongField( env, obj, field, info->lFree );

	field = (*env)->GetFieldID( env, cls, "availSize", "J" );
	(*env)->SetLongField( env, obj, field, info->lAvail );

	field = (*env)->GetFieldID( env, cls, "usedSize", "J" );
	(*env)->SetLongField( env, obj, field, info->lUsed );
}

/*
 * Class:     jp_co_sds_0005fcorp_util_device_DeviceManager
 * Method:    getDriveInfo
 * Signature: (Ljava/lang/String;)Ljp/co/sds_corp/util/device/DriveInfo;
 */
JNIEXPORT jobject JNICALL Java_jp_co_sds_1corp_util_device_DeviceManager_getDriveInfo
  (JNIEnv *env, jclass cls, jstring name)
{
	int		nRet;
	jclass objCls;
	jobject obj;
	struct driveinfo	info;
	const char *pName = (*env)->GetStringUTFChars( env, name, 0 );

	objCls = (*env)->FindClass( env, "jp/co/sds_corp/util/device/DriveInfo" );
	obj = (*env)->AllocObject( env, objCls );

	memset( &info, 0, sizeof( info ));
	nRet = getDriveInfo( env, pName, &info );
	if ( nRet < 0 ) {
		return NULL;
	}
	nRet = getDeviceInfo( env, pName, &info );
	setInfo( env, objCls, obj, &info );

	return obj;
}

#if defined( HAVE_STATVFS )
int getDriveInfo( JNIEnv *env, const char *pName, struct driveinfo *info )
{
	int			nRet;
	struct statvfs stat;

	nRet = statvfs( pName, &stat );
	if ( nRet != 0 ) {
		Throw( env, "java/io/FileNotFoundException", pName );
		return -1;
	}

#ifdef HAVE_F_V_MNTONNAME
	strcpy( info->name, stat.f_mntonname );
	strcpy( info->dev, stat.f_mntfromname );
	strcpy( info->type, stat.f_fstypename );
#else
	strcpy( info->name, pName );
#endif
#ifdef HAVE_F_V_BASETYPE
	strcpy( info->type, stat.f_basetype );
#endif
	info->lBlock = (long long )stat.f_frsize;
	info->lMax = info->lBlock * (long long )stat.f_blocks;
	info->lFree  = info->lBlock * (long long )stat.f_bfree;
	info->lAvail  = info->lBlock * (long long )stat.f_bavail;
	info->lUsed = info->lMax - info->lFree;

	return 0;
}
#elif defined(HAVE_STATFS)
int getDriveInfo( JNIEnv *env, const char *pName, struct driveinfo *info )
{
	int			nRet;

	struct statfs stat;
	nRet = statfs( pName, &stat );
	if ( nRet != 0 ) {
		Throw( env, "java/io/FileNotFoundException", pName );
		return -1;
	}

#ifdef HAVE_F_MNTONNAME
	strcpy( info->name, stat.f_mntonname );
	strcpy( info->dev, stat.f_mntfromname );
	strcpy( info->type, stat.f_fstypename );
#else
	strcpy( info->name, pName );
#endif
	info->lBlock = (long long )stat.f_bsize;
	info->lMax = info->lBlock * (long long )stat.f_blocks;
	info->lFree  = info->lBlock * (long long )stat.f_bfree;
	info->lAvail  = info->lBlock * (long long )stat.f_bavail;
	info->lUsed = info->lMax - info->lFree;

	return 0;
}

#endif

