#ifdef HAVE_CONFIG_H
#include	"config.h"
#endif

#include	<stdio.h>
#ifdef HAVE_STDLIB_H
#include	<stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include	<string.h>
#endif

#ifdef HAVE_KSTAT_H
#include    <kstat.h>
#endif
#ifdef HAVE_SYS_SYSINFO_H
#include    <sys/sysinfo.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include	<sys/types.h>
#endif
#ifdef HAVE_SYS_SYSCTL_H
#include	<sys/sysctl.h>
#endif
#ifdef HAVE_SYS_VMMETER_H
#include	<sys/vmmeter.h>
#endif
#ifdef HAVE_MACH_MACH_H
#include	<mach/mach.h>
#endif
#ifdef HAVE_MACH_HOST_INFO_H
#include	<mach/host_info.h>
#endif

#ifdef HAVE_W32API_BASETYPS_H
#include	<w32api/basetyps.h>
#endif
#include <jni.h>

#include "jp_co_sds_0005fcorp_util_device_DeviceManager.h"

void Throw( JNIEnv *env, const char *name, const char *msg);

struct memusage {
	long long    maxSize;
	long long    usedSize;
	long long    freeSize;
	long long    cacheSize;
	long long    maxSwapSize;
	long long    usedSwapSize;
	long long    freeSwapSize;
};

static void readMeminfo( JNIEnv *env, struct memusage *info );
static void readKstat( JNIEnv *env, struct memusage *memusage );
static void readSysctl( JNIEnv *env, struct memusage *memusage );

#ifdef HAVE_LIBKSTAT
long long getKStatNumber( JNIEnv *env, kstat_ctl_t *kernelDesc, char *moduleName,
			char *recordName, char *fieldName );
char *getKStatString( JNIEnv *env, kstat_ctl_t *kernelDesc, char *moduleName,
			char *recordName, char *fieldName );
#endif
#ifdef HAVE_SYSCTL
void getSysctlString( int n1, int n2, char *buf, size_t len );
int getSysctlInt( int n1, int n2 );
#endif

/*
 * Class:     jp_co_sds_0005fcorp_util_device_DeviceManager
 * Method:    getMemoryUsage
 * Signature: ()Ljp/co/sds_corp/util/device/MemoryUsage;
 */
JNIEXPORT jobject JNICALL Java_jp_co_sds_1corp_util_device_DeviceManager_getMemoryUsage
	(JNIEnv *env, jclass cls)
{
	jfieldID    field;
	struct memusage memusage;
	jclass 		retCls;
	jobject ret;

	retCls = (*env)->FindClass( env, "jp/co/sds_corp/util/device/MemoryUsage" );
	ret = (*env)->AllocObject( env, retCls );

	memset( &memusage, 0, sizeof( struct memusage ));
#ifdef HAVE_LIBKSTAT
	readKstat( env, &memusage );
#else
#ifdef HAVE_PROC_MEMINFO
	readMeminfo( env, &memusage );
#else
	readSysctl( env, &memusage );
#endif
#endif
	field = (*env)->GetFieldID( env, retCls, "maxSize", "J" );
	(*env)->SetLongField( env, ret, field, memusage.maxSize );
	field = (*env)->GetFieldID( env, retCls, "usedSize", "J" );
	(*env)->SetLongField( env, ret, field, memusage.usedSize );
	field = (*env)->GetFieldID( env, retCls, "freeSize", "J" );
	(*env)->SetLongField( env, ret, field, memusage.freeSize );
	field = (*env)->GetFieldID( env, retCls, "cacheSize", "J" );
	(*env)->SetLongField( env, ret, field, memusage.cacheSize );
	field = (*env)->GetFieldID( env, retCls, "maxSwapSize", "J" );
	(*env)->SetLongField( env, ret, field, memusage.maxSwapSize );
	field = (*env)->GetFieldID( env, retCls, "usedSwapSize", "J" );
	(*env)->SetLongField( env, ret, field, memusage.usedSwapSize );
	field = (*env)->GetFieldID( env, retCls, "freeSwapSize", "J" );
	(*env)->SetLongField( env, ret, field, memusage.freeSwapSize );

	return ret;
}

static void readMeminfo( JNIEnv *env, struct memusage *info ) {
	char	*file;
	FILE	*pFp;
	char	buf[1024];
	char	*key;
	int		nCpu = -1;

	file = "/proc/meminfo";
	pFp = fopen( file, "r" );
	if ( pFp == NULL ) {
		Throw( env, "java/io/FileNotFoundException", file );
		return;
	}

	for ( ;; ) {
		if ( fgets( buf, sizeof( buf ), pFp ) == NULL ) {
			break;
		}
		key = "Mem:";
		if ( memcmp( buf, key, strlen( key )) == 0 ) {
			int nPos = 0;
			char *p;
			p = strtok( buf, " " );
			for ( ;; ) {
				if ( p[strlen(p)-1] == '\n' ) {
					p[strlen(p)-1] = '\0';
				}
				switch ( nPos ) {
				case 1: info->maxSize = atoll( p ); break;
				case 2: info->usedSize = atoll( p ); break;
				case 3: info->freeSize = atoll( p ); break;
				case 6: info->cacheSize = atoll( p ); break;
				}
				p = strtok( NULL, " " );
				if ( p == NULL ) {
					break;
				}
				nPos++;
			}
		}
		key = "Swap:";
		if ( memcmp( buf, key, strlen( key )) == 0 ) {
			int nPos = 0;
			char *p;
			p = strtok( buf, " " );
			for ( ;; ) {
				if ( p[strlen(p)-1] == '\n' ) {
					p[strlen(p)-1] = '\0';
				}
				switch ( nPos ) {
				case 1: info->maxSwapSize = atoll( p ); break;
				case 2: info->usedSwapSize = atoll( p ); break;
				case 3: info->freeSwapSize = atoll( p ); break;
				}
				p = strtok( NULL, " " );
				if ( p == NULL ) {
					break;
				}
				nPos++;
			}
		}
	}
	fclose( pFp );
}

#ifdef HAVE_LIBKSTAT
#define	BYTE_PAGE		8192

extern kstat_ctl_t *kc;
extern initKstat();

static void readKstat( JNIEnv *env, struct memusage *memusage )
{
	long long	value;
	char	*str;
	kstat_t		*ksp;
	vminfo_t	vm1, vm2;

	initKstat();
	/* Fetch information on physical memory usage */
	//value = getKStatNumber( env, kc, "unix", "system_pages", "physmem");
	value = getKStatNumber( env, kc, "unix", "system_pages", "availrmem");
	memusage->maxSize = value * (long long )BYTE_PAGE;

	value = getKStatNumber( env, kc, "unix", "system_pages", "freemem");
	memusage->freeSize = value * (long long )BYTE_PAGE;
	memusage->usedSize = memusage->maxSize - memusage->freeSize;

	ksp = kstat_lookup( kc, "unix", 0, "vminfo" );
	if ( ksp == NULL ) {
		perror( "kstat_lookup" );
		return;
	}
	if ( kstat_read( kc, ksp, &vm1 ) == -1 ) {
		perror( "kstat_read" );
		return;
	}
	sleep( 1 );
	if ( kstat_read( kc, ksp, &vm2 ) == -1 ) {
		perror( "kstat_read" );
		return;
	}
	memusage->maxSwapSize
		= ( vm2.swap_resv - vm1.swap_resv + vm2.swap_avail - vm1.swap_avail )
			* BYTE_PAGE;
	memusage->usedSwapSize
		= ( vm2.swap_resv - vm1.swap_resv ) * BYTE_PAGE;
	memusage->freeSwapSize = memusage->maxSwapSize - memusage->usedSwapSize;
}
#endif

#if defined(HAVE_SYSCTL) && !defined(HAVE_PROC_MEMINFO)
static void printSysVmmeter() {
	int mib[2];
	size_t len;
	struct vmtotal vm;

	mib[0] = CTL_VM;
	mib[1] = VM_METER;
	sysctl(mib, 2, &vm, &len, NULL, 0);
/*
	printf( "total vmemory=%llu\n", vm.t_vm );
	printf( "active vmemory=%llu\n", vm.t_avm );
	printf( "total rmemory=%llu\n", vm.t_rm );
	printf( "active rmemory=%llu\n", vm.t_arm );
	printf( "free memory=%llu\n", vm.t_free );
*/
}

static void readSysctl( JNIEnv *env, struct memusage *memusage )
{
	int	val;
	int	pagesize;
	int i;
	unsigned int count = HOST_VM_INFO_COUNT;
	vm_statistics_data_t vm_stats;
	long cp_time[CPU_STATE_MAX];

	pagesize = getSysctlInt( CTL_HW, HW_PAGESIZE );
	val = getSysctlInt( CTL_HW, HW_PHYSMEM );
	memusage->maxSize = val;

	host_statistics(mach_host_self(), HOST_VM_INFO,
					(host_info_t)&vm_stats, &count);
	memusage->usedSize = ( vm_stats.active_count + vm_stats.inactive_count
					+ vm_stats.wire_count ) * pagesize;
	memusage->freeSize = vm_stats.free_count * pagesize;
}
#endif
